package lsedit;

import java.awt.Color;
import java.util.Vector;
import java.io.IOException;
import java.io.PrintStream;

import javax.swing.undo.UndoableEdit;

public class EntityClass extends LandscapeClassObject /* extends LandscapeObject3D extends LandscapeObject */ 
{
	class UpdateImage extends MyUndoableEdit implements UndoableEdit
	{
		EntityClass			 m_o;
		int                  m_old;
		int                  m_new;

		UpdateImage(EntityClass o, int value)
		{
			m_o   = o;
			m_old = o.getImage();
			m_new = value;
			if (logEdit(this)) {
				o.setImage(value);
		}	}

		public String getPresentationName() 
		{
			return m_o + " Change Image";
		}

		public void undo()
		{
			m_o.setImage(m_old);
		}

		public void redo()
		{
			m_o.setImage(m_new);
	}	}

	class UpdateAngle extends MyUndoableEdit implements UndoableEdit
	{
		EntityClass			 m_o;
		double               m_old;
		double               m_new;

		UpdateAngle(EntityClass o, double value)
		{
			m_o   = o;
			m_old = o.getAngle();
			m_new = value;
			if (logEdit(this)) {
				o.setAngle(value);
		}	}

		public String getPresentationName() 
		{
			return m_o + " Angle " + m_new;
		}

		public void undo()
		{
			m_o.setAngle(m_old);
		}

		public void redo()
		{
			m_o.setAngle(m_new);
	}	}

	public static final String ENTITY_BASE_CLASS_ID		= "$ENTITY";

	public static final int ENTITY_STYLE_NONE           = -1;
	public static final int ENTITY_STYLE_3DBOX			= 0;
	public static final int ENTITY_STYLE_2DBOX			= 1;
	public static final int ENTITY_STYLE_FILE			= 2;
	public static final int ENTITY_STYLE_DISK			= 3;
	public static final int ENTITY_STYLE_FOLDER			= 4;
	public static final int ENTITY_STYLE_SOURCEOBJ		= 5;
	public static final int ENTITY_STYLE_CLASS			= 6;
	public static final int ENTITY_STYLE_GROUP			= 7;
	public static final int ENTITY_STYLE_LABELLED_GROUP = 8;
	public static final int ENTITY_STYLE_TRIANGLE       = 9;
	public static final int ENTITY_STYLE_ROMBUS		    = 10;
	public static final int ENTITY_STYLE_TRAPEZOID      = 11;
	public static final int ENTITY_STYLE_TRIANGLE2      = 12;

	public static final int ENTITY_STYLE_RECTANGLE      = 13;
	public static final int ENTITY_STYLE_PENTAGON       = 14;
	public static final int ENTITY_STYLE_HEXAGON        = 15;
	public static final int ENTITY_STYLE_OCTAGON        = 16;
	public static final int ENTITY_STYLE_DECAHEDRON     = 17;
	public static final int ENTITY_STYLE_12SIDED        = 18;
	public static final int ENTITY_STYLE_14SIDED        = 19;
	public static final int ENTITY_STYLE_16SIDED        = 20;
	public static final int ENTITY_STYLE_18SIDED        = 21;
	public static final int ENTITY_STYLE_20SIDED        = 22;
	public static final int ENTITY_STYLE_PAPER          = 23;
	public static final int ENTITY_STYLE_SUNK_BOX       = 24;


	public static final String[] styleName =
		{ 
				"3D Box", 
				"2D Box", 
				"File", 
				"Disk", 
				"Folder", 
				"Source Object", 
				"Class", 
				"Group", 
				"Labelled Group",
				"Triangle  (directed)",
				"Rombus    (directed)",
				"Trapezoid (directed)",
				"3-sided  Triangle",
				"4-sided  Diamond",
				"5-sided  Pentagon",
				"6-sided  Hexagon",
				"8-sided  Octagon",
				"10-sided Decahedron",
				"12 sided Polygon",
				"14 sided Polygon",
				"16 sided Polygon",
				"18 sided Polygon",
				"20 sided Dodecahedron",
				"Paper",
				"Sunken Box"
		};


	public static final int ENTITY_IMAGE_NONE           =   0x00;
	public static final int ENTITY_IMAGE_ACTOR          =   0x01;
	public static final int ENTITY_IMAGE_OVAL           =   0x02;
	public static final int ENTITY_IMAGE_FRAME          =   0x04;
	public static final int ENTITY_IMAGE_ROUNDED_FRAME  =   0x08;
	public static final int ENTITY_IMAGE_X              =   0x10;

	public static final int ENTITY_IMAGE_LAST           =   ENTITY_IMAGE_X;

	public static final String[] imageName =
		{ 
				"Actor",
				"Oval",
				"Frame",
				"Rounded Frame",
				"X"
		};

	private boolean	m_shown = true;
	private	int		m_orderedId;

	// ----------------
	// Exported methods
	// ----------------

	// Constructor 

	public EntityClass(String id, int nid, Diagram dg) 
	{
		setId(id);
		setNid(nid); 
		setLabel(id); 
		setLandscapeObjectDiagram(dg);
	}

	public void setOrderedId(int orderedId)
	{
		m_orderedId = orderedId;
	}

	public int getOrderedId()
	{
		return(m_orderedId);
	}

	public static String getEntityStyleName(int style) 
	{
		if (style < 0 || style >= EntityClass.styleName.length) {
			return "";
		}
		return styleName[style];
	}

	public String getStyleName(int style)
	{
		return EntityClass.getEntityStyleName(style);
	}

	public void updateImage(int value)
	{
		if (value != m_image) {
			new UpdateImage(this, value);
	}	}

	public void updateAngle(double value)
	{
		if (value != m_angle) {
			new UpdateAngle(this, value);
	}	}

	public void setShown(boolean value)
	{
		if (m_shown != value) {
			if (!value) {
				Vector grp = getDiagram().getGroup();

				if (grp != null) {
					int				i;
					EntityInstance	e;
					EntityClass		ec;
					boolean			seenKeyEntry = false;

					for (i = grp.size(); --i >= 0; ) {
						e  = (EntityInstance) grp.elementAt(i);
						ec = e.getEntityClass();
						if (ec == this) {
							e.clearGroupFlag();
							grp.removeElementAt(i);
						} else {
							seenKeyEntry |= e.getGroupKeyFlag();
					}	}
					if (!seenKeyEntry && grp.size() > 0) {
						e = (EntityInstance) grp.elementAt(0);
						e.setGroupKeyFlag();
			}	}	}
			m_shown = value;
	}	}

	public boolean isShown()
	{
		return(m_shown);
	}

	public void writeAttributes(PrintStream ps) throws IOException {

		ps.print(qt(getId()) + " {\n");
		super.writeAttributes(ps);
		ps.print("}\n\n"); 
	}

	public EntityInstance newEntity(String id) 
	{
		EntityInstance e = new EntityInstance(this, id);
		return e;
	}

	public void shapeChanges(int oldStyle, double oldAngle)
	{
		Diagram			diagram;
		EntityInstance	root;
		int				style;

		style = getInheritedStyle();
		if (oldStyle == style) {
			if (style < ENTITY_STYLE_TRIANGLE || style > ENTITY_STYLE_20SIDED) {
				// Changing the angle is irrelevant for this style
				return;
		}	}

		diagram = getDiagram();
		if (diagram == null) {
			return;
		}

		root = diagram.getRootInstance();
		if (root == null) {
			return;
		}
		root.shapeChanges(this);
	}

	// The routines that follow hide the complexity of getting/setting attribute values
	// from EditAttributes

	public int getPrimaryAttributeCount()
	{
		return(9);
	}

	public String getLsAttributeNameAt(int index)
	{
		String	name;

		switch (index) {
		case 0:
			name  = "id";
			break;
		case 1:
			name  = CLASSLABEL_ID;
			break;
		case 2:
			name  = CLASSDESC_ID;
			break;
		case 3:
			name  = CLASSSTYLE_ID;
			break;
		case 4:
			name  = CLASSANGLE_ID;
			break;
		case 5:	
			name  = CLASSIMAGE_ID;
			break;
		case 6:
			name  = COLOUR_ID;
			break;
		case 7:
			name  = LABEL_COLOUR_ID;
			break;
		case 8:
			name  = OPEN_COLOUR_ID;
			break;
		default:
			name  = super.getLsAttributeNameAt(index);
		}
		return(name);
	}

	public Object getLsAttributeValueAt(int index)
	{
		Object	value;

		switch (index) {
		case 0:
			value = getId();
			break;
		case 1:
			value = getLabel();
			break;
		case 2:
			value = getDescription();
			break;
		case 3:
			value = new Integer(getStyle());
			break;
		case 4:
			value = new Double(getAngle());
			break;
		case 5:
			value = new Integer(getImage());
			break;
		case 6:
			value = getObjectColor();
			break;
		case 7:
			value = getLabelColor();
			break;
		case 8:
			value = getColorWhenOpen();
			break;
		default:
			value = super.getLsAttributeValueAt(index);
		}
		return(value);
	}

	public void setAttributeValueAt(int index, Object value)
	{
		switch (index) {
		case 0:
			setId((String) value);
			break;
		case 1:
			updateLabel((String) value);
			break;
		case 2:
			updateDescription((String) value);
			break;
		case 3:
			updateStyle(((Integer) value).intValue());
			break;
		case 4:
			updateAngle(((Double) value).doubleValue());
			break;
		case 5:
			updateImage(((Integer) value).intValue());
			break;
		case 6:
			updateObjectColor((Color) value);
			break;
		case 7:
			updateLabelColor((Color) value);
			break;
		case 8:
			updateColorWhenOpen((Color) value);
			break;
		default:
			super.setAttributeValueAt(index, value);
	}	}

	// Need to know the type in cases where value might be null
	// For example with some colors

	public int getLsAttributeTypeAt(int index)
	{
		int		ret;
		
		switch (index) {
		case 0:
		case 1:
			ret = Attribute.STRING_TYPE;
			break;
		case 2:
			ret = Attribute.TEXT_TYPE;
			break;
		case 3:
			ret = Attribute.ENTITY_STYLE_TYPE;
			break;
		case 4:
			ret = Attribute.DOUBLE_TYPE;
			break;
		case 5:
			ret = Attribute.ENTITY_IMAGE_TYPE;
			break;
		case 6:
		case 7:
		case 8:
			ret = Attribute.COLOR_OR_NULL_TYPE;
			break;
		default:
			ret = super.getLsAttributeTypeAt(index);
		}
		return(ret);
	}
}



